"""
This module contains base classes (some of which are abstract) common to both **Source** and
**Ground Motion Model (GMM)** logic trees.
"""
import copy
import json
import math
from abc import ABC, abstractmethod
from collections.abc import Sequence
from dataclasses import asdict, dataclass, field, fields
from functools import reduce
from itertools import product
from operator import mul
from pathlib import Path
from typing import Any, Dict, Generator, Iterator, List, Type, Union

import dacite

from nzshm_model.psha_adapter import PshaAdapterInterface

from .branch import Branch, CompositeBranch
from .correlation import LogicTreeCorrelations

# TODO:
# - move values to the base class?
# - move logic for creating light branchSet and LogicTree when creating FilteredBranchs to
#    the initializer of FilteredBranch?
# - FilteredBranch doesn't need to be a data class as it should not be serialized and doesn't contain many arguments
# - should we use FilteredBranch for correlation so the branches can be traced back to the BranchSet?


@dataclass
class BranchSet:
    """
    A group of branches that comprise their own sub-logic tree. Also known as a fault system logic
    tree (for source logic trees).

    Arguments:
        short_name:
        long_name:
        branches: the branches that make up the branch set
    """

    short_name: str = ''
    long_name: str = ''
    branches: Sequence[Any] = field(default_factory=list)

    def __post_init__(self):
        if not self._validate_weights():
            raise ValueError("weights of BranchSet must sum to 1.0")

    def _validate_weights(self) -> bool:
        """
        verify that weighs sum to 1.0
        """
        weight = 0.0
        if not self.branches:  # empty BranchSet
            return True
        for b in self.branches:
            weight += b.weight
        return math.isclose(weight, 1.0)


@dataclass
class LogicTree(ABC):
    """
    Logic tree baseclass. Contains information about branch sets and correlations between branches of the branch sets.

    The correlations are enforced when forming composite branches (combinations of branches from different branch sets).

    Arguments:
        title: title of the logic tree
        version: version string
        branch_sets: list of branch sets that make up the logic tree
        correlations: any correlations between branches of the branch_sets

    """

    title: str = ''
    version: str = ''
    branch_sets: List[Any] = field(default_factory=list)
    correlations: LogicTreeCorrelations = field(default_factory=LogicTreeCorrelations)

    def __post_init__(self) -> None:
        self._validate_correlations()

    def __setattr__(self, __name: str, __value: Any) -> None:
        super().__setattr__(__name, __value)
        if __name == "correlations":
            self._validate_correlations()

    def _validate_correlations(self) -> None:
        # check that the weights total 1.0
        weight_total = 0.0
        for branch in self.combined_branches:
            weight_total += branch.weight
        if not math.isclose(weight_total, 1.0):
            raise ValueError("the weights of the logic tree do not sum to 1.0 when correlations are applied")

    def _combined_branches(self) -> Generator[CompositeBranch, None, None]:
        """
        yields all composite (combined) branches of the branch_sets without applying correlations.
        """
        for branches in product(*[branch_set.branches for branch_set in self.branch_sets]):
            yield CompositeBranch(branches=branches)

    @property
    def combined_branches(self) -> Generator[CompositeBranch, None, None]:
        """
        yields all composite (combined) branches of the branch_sets enforcing correlations
        """
        for combined_branch in self._combined_branches():
            # if the comp_branch contains a branch listed as the 0th element of the correlations, only
            # yeild if the other branches are present
            # weight is automatically calculated by CompositeBranch if not explicitly assigned
            # (as we would with correlations)
            correlation_match = [branch_pri in combined_branch for branch_pri in self.correlations.primary_branches()]
            if any(correlation_match):
                i_cor = correlation_match.index(
                    True
                )  # index of the correlation that matches a branch in _combined_branches()
                if not all(br in combined_branch for br in self.correlations[i_cor].all_branches):
                    continue
                weights = [self.correlations[i_cor].weight] + [
                    branch.weight for branch in combined_branch if branch not in self.correlations[i_cor].all_branches
                ]
                combined_branch.weight = reduce(mul, weights, 1.0)
            yield combined_branch

    @classmethod
    def from_json(cls, json_path: Union[Path, str]) -> 'LogicTree':
        """
        Create LogicTree object from json file or string

        Parameters:
            json_path: path to json file

        Returns:
            logic_tree
        """
        with Path(json_path).open() as jsonfile:
            data = json.load(jsonfile)
        return cls.from_dict(data)

    @classmethod
    def from_dict(cls, data: Dict) -> 'LogicTree':
        """
        Create LogicTree object from dict

        Parameters:
            data: dict representation of LogicTree object

        Returns:
            logic_tree
        """
        return dacite.from_dict(data_class=cls, data=data, config=dacite.Config(strict=True))

    def to_dict(self) -> Dict[str, Any]:
        return asdict(self)

    # would like this to actully do the work, but not sure how to pass the logic trees wihtout knowning the type.
    # Could check for type in PshaAdaptorInterface, but then we have a circular import.
    @abstractmethod
    def psha_adapter(self, provider: Type[PshaAdapterInterface], **kwargs):
        """
        Provide an adapter object for translating LogicTrees to/from specific formats

        Parameters:
            provider: the interface object that defines a specific implimenation
            **kwargs:

        Returns:
            An adapter object
        """
        pass

    def __all_branches__(self) -> Generator['FilteredBranch', None, None]:
        """
        Yield all branches from all BranchSets, each with a shallow copy of its LogicTree and BranchSet parents
        for use in filtering.

        NB this class is never used for serialising models.
        """

        def get_fields(obj):
            return {
                field.name: copy.deepcopy(getattr(obj, field.name))
                for field in fields(obj)
                if field.name not in ('branches', 'branch_sets')
            }

        lt_fields = get_fields(self)
        for branch_set in self.branch_sets:
            bs_fields = get_fields(branch_set)
            for branch in branch_set.branches:
                bs_lite = type(branch_set)(**bs_fields, branches=[type(branch)()])
                lt_lite = type(self)(**lt_fields)
                # b_fields = get_fields(branch)
                yield branch.filtered_branch(
                    branch_set=bs_lite,
                    logic_tree=lt_lite,
                )

    @classmethod
    def from_branches(cls, branches: Iterator['FilteredBranch']) -> 'LogicTree':
        """
        Build a complete LogicTree from a iterable of branches.

        We expect that all the branches have come from a single LogicTree.

        Parameters:
            branches: the branches used to build the LogicTree
        """

        def match_branch_set(slt: LogicTree, fb):
            for branch_set in slt.branch_sets:
                if fb.branch_set.short_name == branch_set.short_name:
                    return branch_set

        version = None
        for fb in branches:
            # ensure an slt
            if version is None:
                logic_tree = cls(version=fb.logic_tree.version, title=fb.logic_tree.title)
                version = fb.logic_tree.version
            else:
                assert version == fb.logic_tree.version

            # ensure an branch_set
            bs = match_branch_set(logic_tree, fb)
            if not bs:
                bs = type(fb.branch_set)(short_name=fb.branch_set.short_name, long_name=fb.branch_set.long_name)
                logic_tree.branch_sets.append(bs)
            bs.branches.append(fb.to_branch())
        return logic_tree

    def __iter__(self):
        self.__current_branch = 0
        self.__branch_list = list(self.__all_branches__())
        return self

    def __next__(self):
        if self.__current_branch >= len(self.__branch_list):
            raise StopIteration
        else:
            self.__current_branch += 1
            return self.__branch_list[self.__current_branch - 1]


@dataclass
class FilteredBranch(Branch):
    """
    A branch type that points back to it's logic tree and branch set. Should never be serialized, only
    used for filtering
    """

    logic_tree: LogicTree = field(default_factory=LogicTree)
    branch_set: BranchSet = field(default_factory=BranchSet)

    def to_branch(self) -> Branch:
        """
        Produce the Branch object (does not point back to it's logic tree and branch set)

        Returns:
            branch: the Branch object
        """
        branch_attributes = {k: v for k, v in self.__dict__.items() if k not in ('branch_set', 'logic_tree')}
        return type(self.branch_set.branches[0])(**branch_attributes)



gt_description = "Source Logic Tree v8"


logic_tree_permutations = [
    [
        {
            "tag": "Final SLT for TAG Workshop",
            "weight": 1.0,
            "permute": [
                {
                    "group": "Hik",
                    "members": [
                        {
                            "tag": "Hik TL, N16.5, b0.95, C4, s1",
                            "weight": 0.4,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwMzM5MjI=",
                            "bg_id": "RmlsZToxMzA3MzQ=",
                        },
                        {
                            "tag": "Hik TL, N21.5, b1.097, C4, s1",
                            "weight": 0.6,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwMzM5MjE=",
                            "bg_id": "RmlsZToxMzA3NDA=",
                        },
                    ],
                },
                {
                    "group": "PUY",
                    "members": [
                        {
                            "tag": "Puy 0.7, N4.6, b0.902, C4, s1",
                            "weight": 1.0,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjExODcyMw==",
                            "bg_id": "RmlsZToxMzA3NTQ=",
                        },
                    ],
                },
                {
                    "group": "CRU",
                    "members": [
                        {
                            "tag": "geodetic, TI, N2.7, b0.823 C4.2 s1",
                            "weight": 0.35,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEyMDkxMQ==",
                            "bg_id": "RmlsZToxMzA3MTA=",
                        },
                        {
                            "tag": "geodetic, TI, N3.4, b0.959 C4.2 s1",
                            "weight": 0.65,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEyMDkwNg==",
                            "bg_id": "RmlsZToxMzA3MTk=",
                        },
                    ],
                },
                {
                    "group": "SLAB",
                    "members": [
                        {"tag": "slab-uniform-1depth-rates", "weight": 1.0, "inv_id": "", "bg_id": "RmlsZToxMjEwMzM="}
                    ],
                },
            ],
        }
    ]
]



gt_description = "Sample Large Source Logic Tree "

logic_tree_permutations = [
    [
        {
            "tag": "Geologic",
            "weight": 1.0,
            "permute": [
                {
                    "group": "Hik",
                    "members": [
                        {
                            "tag": "hiktlck, b0.979, C3.9, s0.78",
                            "weight": 0.0666666666666667,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODA3NQ==",
                            "bg_id": "RmlsZToxMDY1MjU=",
                        },
                        {
                            "tag": "hiktlck, b0.979, C3.9, s1",
                            "weight": 0.0666666666666667,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODA4Mw==",
                            "bg_id": "RmlsZToxMDY1MzA=",
                        },
                        {
                            "tag": "hiktlck, b0.979, C3.9, s1.22",
                            "weight": 0.0666666666666667,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODA5MA==",
                            "bg_id": "RmlsZToxMDY1Mjg=",
                        },
                        {
                            "tag": "hiktlck, b0.979, C3.9, s1.87",
                            "weight": 0.0666666666666667,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODA3OA==",
                            "bg_id": "RmlsZToxMDY1Mjg=",
                        },
                        {
                            "tag": "hiktcrp, b0.979, C3.9, s1",
                            "weight": 0.0666666666666667,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODA5NQ==",
                            "bg_id": "RmlsZToxMDY1Mjk=",
                        },
                        {
                            "tag": "hiktcrp, b1.101, C4.1, s1",
                            "weight": 0.0666666666666667,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODE5Ng==",
                            "bg_id": "RmlsZToxMDY1NDA=",
                        },
                        {
                            "tag": "hiktcrp, b1.101, C4.1, s0.78",
                            "weight": 0.0666666666666667,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODE5MQ==",
                            "bg_id": "RmlsZToxMDY1MzM=",
                        },
                        {
                            "tag": "hiktcrp, b1.101, C4.1, s1.22",
                            "weight": 0.0666666666666667,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODIwMA==",
                            "bg_id": "RmlsZToxMDY1Mzg=",
                        },
                        {
                            "tag": "hiktcrp, b1.101, C4.1, s1.87",
                            "weight": 0.0666666666666667,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODE5Mg==",
                            "bg_id": "RmlsZToxMDY1Mzg=",
                        },
                        {
                            "tag": "hiktlck, b1.101, C4.1, s1",
                            "weight": 0.0666666666666667,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODEwMw==",
                            "bg_id": "RmlsZToxMDY1Mzk=",
                        },
                        {
                            "tag": "hiktcrp, b1.211, C4.1, s0.78",
                            "weight": 0.0666666666666667,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODE5Nw==",
                            "bg_id": "RmlsZToxMDY1MzE=",
                        },
                        {
                            "tag": "hiktcrp, b1.211, C4.1, s1",
                            "weight": 0.0666666666666667,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODE5OQ==",
                            "bg_id": "RmlsZToxMDY1MjQ=",
                        },
                        {
                            "tag": "hiktcrp, b1.211, C4.1, s1.22",
                            "weight": 0.0666666666666667,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODE5MA==",
                            "bg_id": "RmlsZToxMDY1MzY=",
                        },
                        {
                            "tag": "hiktcrp, b1.211, C4.1, s1.87",
                            "weight": 0.0666666666666667,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODE5Mw==",
                            "bg_id": "RmlsZToxMDY1MzY=",
                        },
                        {
                            "tag": "hiktcrp, b1.211, C4.1, s1",
                            "weight": 0.0666666666666667,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODE5OQ==",
                            "bg_id": "RmlsZToxMDY1MjQ=",
                        },
                    ],
                },
                {
                    "group": "PUY",
                    "members": [
                        {
                            "tag": "puy, b0.791, C3.9, s0.57",
                            "weight": 0.111111111111111,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODI3MA==",
                            "bg_id": "RmlsZToxMDY1NDg=",
                        },
                        {
                            "tag": "puy, b0.791, C3.9, s1",
                            "weight": 0.111111111111111,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODI4NQ==",
                            "bg_id": "RmlsZToxMDY1NTQ=",
                        },
                        {
                            "tag": "puy, b0.791, C3.9, s1.43",
                            "weight": 0.111111111111111,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODI3OA==",
                            "bg_id": "RmlsZToxMDY1NDE=",
                        },
                        {
                            "tag": "puy, b0.882, C4, s1",
                            "weight": 0.111111111111111,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODIzOQ==",
                            "bg_id": "RmlsZToxMDY1NzA=",
                        },
                        {
                            "tag": "puy, b0.882, C4, s0.57",
                            "weight": 0.111111111111111,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODI0Ng==",
                            "bg_id": "RmlsZToxMDY1NjE=",
                        },
                        {
                            "tag": "puy, b0.882, C4, s1.43",
                            "weight": 0.111111111111111,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODI0OQ==",
                            "bg_id": "RmlsZToxMDY1NTI=",
                        },
                        {
                            "tag": "puy, b1.064, C4.1, s0.57",
                            "weight": 0.111111111111111,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODI5Mw==",
                            "bg_id": "RmlsZToxMDY1NTg=",
                        },
                        {
                            "tag": "puy, b1.064, C4.1, s1",
                            "weight": 0.111111111111111,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODI4Mg==",
                            "bg_id": "RmlsZToxMDY1NDQ=",
                        },
                        {
                            "tag": "puy, b1.064, C4.1, s1.43",
                            "weight": 0.111111111111111,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODI3NQ==",
                            "bg_id": "RmlsZToxMDY1NjU=",
                        },
                    ],
                },
                {
                    "group": "CRU",
                    "members": [
                        {
                            "tag": "Cru_geol, b0.849, C4.1, s0.53",
                            "weight": 0.0333333333333333,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODMzNw==",
                            "bg_id": "RmlsZToxMDY1NDc=",
                        },
                        {
                            "tag": "Cru_geol, b0.849, C4.1, s0.66",
                            "weight": 0.0333333333333333,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODM0MQ==",
                            "bg_id": "RmlsZToxMDY1Njg=",
                        },
                        {
                            "tag": "Cru_geol, b0.849, C4.1, s1",
                            "weight": 0.0333333333333333,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODMzNA==",
                            "bg_id": "RmlsZToxMDY1NTM=",
                        },
                        {
                            "tag": "Cru_geol, b0.849, C4.1, s1.47",
                            "weight": 0.0333333333333333,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODM0MA==",
                            "bg_id": "RmlsZToxMDY1NDI=",
                        },
                        {
                            "tag": "Cru_geol, b0.849, C4.1, s1.61",
                            "weight": 0.0333333333333333,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODMzMg==",
                            "bg_id": "RmlsZToxMDY1NjQ=",
                        },
                        {
                            "tag": "Cru_geol, b0.952, C4.2, s1",
                            "weight": 0.0333333333333333,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODMyMg==",
                            "bg_id": "RmlsZToxMDY1Njk=",
                        },
                        {
                            "tag": "Cru_geol, b0.952, C4.2, s0.53",
                            "weight": 0.0333333333333333,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODMyNA==",
                            "bg_id": "RmlsZToxMDY1NjI=",
                        },
                        {
                            "tag": "Cru_geol, b0.952, C4.2, s0.66",
                            "weight": 0.0333333333333333,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODMyMA==",
                            "bg_id": "RmlsZToxMDY1NjA=",
                        },
                        {
                            "tag": "Cru_geol, b0.952, C4.2, s1.47",
                            "weight": 0.0333333333333333,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODMyNg==",
                            "bg_id": "RmlsZToxMDY1NTE=",
                        },
                        {
                            "tag": "Cru_geol, b0.952, C4.2, s1.61",
                            "weight": 0.0333333333333333,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODMzMA==",
                            "bg_id": "RmlsZToxMDY1NDk=",
                        },
                        {
                            "tag": "Cru_geol, b1.06, C4.3, s0.53",
                            "weight": 0.0333333333333333,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODM0Ng==",
                            "bg_id": "RmlsZToxMDY1NTc=",
                        },
                        {
                            "tag": "Cru_geol, b1.06, C4.3, s0.66",
                            "weight": 0.0333333333333333,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODM0NQ==",
                            "bg_id": "RmlsZToxMDY1NTY=",
                        },
                        {
                            "tag": "Cru_geol, b1.06, C4.3, s1",
                            "weight": 0.0333333333333333,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODM0OQ==",
                            "bg_id": "RmlsZToxMDY1NDM=",
                        },
                        {
                            "tag": "Cru_geol, b1.06, C4.3, s1.47",
                            "weight": 0.0333333333333333,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODM1MA==",
                            "bg_id": "RmlsZToxMDY1NjY=",
                        },
                        {
                            "tag": "Cru_geol, b1.06, C4.3, s1.61",
                            "weight": 0.0333333333333333,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEwODM0OA==",
                            "bg_id": "RmlsZToxMDY1NDU=",
                        },
                    ],
                },
            ],
        }
    ]
]



import pytest

import nzshm_model as nm

CURRENT_MODEL = 'NSHM_v1.0.4'


@pytest.fixture(scope="module")
def model_104():
    # yield nm.model.NshmModel.get_model_version(CURRENT_MODEL)
    yield nm.get_model_version(CURRENT_MODEL)


def test_list_all_models():
    assert CURRENT_MODEL in nm.all_model_versions()


class TestLoadModel:
    def test_load_model_104(self):
        model = nm.model.NshmModel.get_model_version(CURRENT_MODEL)
        assert model

    def test_model_104_title(self, model_104):
        assert model_104.title == "NSHM version 1.0.4, corrected fault geometry"


class TestGetSourceBranchSets:
    def test_with_list(self, model_104):
        branch_sets = list(model_104.get_source_branch_sets(['CRU', 'PUY']))
        assert len(branch_sets) == 2
        assert branch_sets[0].short_name == 'CRU'
        assert branch_sets[1].short_name == 'PUY'

    def test_with_single(self, model_104):
        branch_sets = list(model_104.get_source_branch_sets('CRU'))
        assert len(branch_sets) == 1
        assert branch_sets[0].short_name == 'CRU'

    def test_with_null(self, model_104):
        assert len(list(model_104.get_source_branch_sets([]))) == 4
        assert len(list(model_104.get_source_branch_sets())) == 4

    def test_with_invalid_branch(self, model_104):
        with pytest.raises(StopIteration):
            next(model_104.get_source_branch_sets(['XXX']))


# class TestGetSourceBranches:
#     def test_get_first_crustal_branch(self, model_104):
#         cru_branches = next(model_104.get_source_branch_sets('CRU')).branches
#         assert next(model_104.get_source_branches('CRU')) == cru_branches[0]

#     def test_invalid_branch_set(self, model_104):
#         with pytest.raises(StopIteration):
#             next(model_104.get_source_branches(['XXX']))
